#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <string>
#include <fmi2Functions.h>
#include <fmi2FunctionTypes.h>
#include <iostream>

using namespace std;

int main(int argc, char **argv) {

    int ret = EXIT_SUCCESS;
    fmi2Real tstart = 0.0;
    fmi2Real tcur = tstart;
    fmi2Real hstep = 0.2;
    fmi2Real tend = 7.0;
    void *handle;
  
    fmi2InstantiateTYPE* (fmi2Instantiate);
    fmi2EnterInitializationModeTYPE* (fmi2EnterInitializationMode);
    fmi2ExitInitializationModeTYPE* (fmi2ExitInitializationMode);
    fmi2SetIntegerTYPE* (fmi2SetInteger);
    fmi2GetIntegerTYPE* (fmi2GetInteger);
    fmi2DoStepTYPE* (fmi2DoStep);
    void (*set_inverted_map)(int,string);
   
    char *error;
    handle = dlopen ("bip_example/binaries/linux64/bip_example.so", RTLD_LAZY);
    if (!handle) {
        fprintf (stderr, "%s\n", dlerror());
        exit(1);
    }
    dlerror();    /* Clear any existing error */
    
    fmi2Instantiate = (fmi2InstantiateTYPE*) dlsym(handle, "fmi2Instantiate");
    fmi2EnterInitializationMode = (fmi2EnterInitializationModeTYPE*) dlsym(handle, "fmi2EnterInitializationMode");
    fmi2ExitInitializationMode = (fmi2ExitInitializationModeTYPE*) dlsym(handle, "fmi2ExitInitializationMode");
    fmi2SetInteger = (fmi2SetIntegerTYPE*) dlsym(handle, "fmi2SetInteger");
    fmi2GetInteger = (fmi2GetIntegerTYPE*) dlsym(handle, "fmi2GetInteger");
    fmi2DoStep = (fmi2DoStepTYPE*) dlsym(handle, "fmi2DoStep");
    set_inverted_map = (void (*)(int, std::string)) dlsym(handle, "set_inverted_map");

    if ((error = dlerror()) != NULL)  {
        fprintf (stderr, "%s\n", error);
        exit(1);
    }
    
    cout<<"CALL fmi2Instantiate...\n";
    
    fmi2Component c = (*fmi2Instantiate)("prModel", fmi2CoSimulation , "prModel", "resource/location", NULL, true, true);
    
    (*set_inverted_map)(100, "ROOT.environment._id__inputIndex");
    //(*set_inverted_map)(101, "ROOT.environment._id__time");
    (*set_inverted_map)(102, "ROOT.camera._id__inputIndex");
    (*set_inverted_map)(103, "ROOT.perception._id__inputIndex");
    (*set_inverted_map)(104, "ROOT.perception._id__result");
    (*set_inverted_map)(105, "ROOT.controller._id__speed");
    (*set_inverted_map)(106, "ROOT.controller._id__speedLimit");
    (*set_inverted_map)(107, "ROOT.controller._id__deltaSpeed");
    (*set_inverted_map)(108, "ROOT.controller._id__zero");
    (*set_inverted_map)(109, "ROOT.brake._id__deltaSpeed");
    (*set_inverted_map)(110, "ROOT.throttle._id__deltaSpeed");
    (*set_inverted_map)(111, "ROOT.speedSensor._id__speed");
    (*set_inverted_map)(112, "ROOT.speedSensor._id__deltaSpeed");
    
    const unsigned int vr[] = {100,/*101,*/ 102,103,104,105,106,107,108,109,110,111,112};
    const int values[] = {1,/*10,*/ 0,0,0,0,0,0,0,0,0,100,0};
    
    cout<<"CALL fmi2SetInteger...\n";
    (*fmi2SetInteger)(c, vr, 12, values);
    
    cout<<"CALL fmi2EnterInitializationMode...\n";
    (*fmi2EnterInitializationMode)(c);
    
    cout<<"CALL fmi2ExitInitializationMode...\n";
    (*fmi2ExitInitializationMode)(c);
    
    tcur = tstart;
    while(tcur<tend){
        fmi2Boolean newStep = fmi2True;
        fmi2Integer val[12];

        cout<<"tcur: "<<tcur<<'\n';
        cout<<"CALL fmi2DoStep...\n";

        fmi2DoStep(c, tcur,hstep, newStep);

        cout<<"CALL fmi2GetInteger...\n";

        fmi2GetInteger(c, vr, 12, val);

        tcur += hstep;
        
     }
    
    cout<<"Simulation finished.\n";
    dlclose(handle);

    return ret;
}
